"""Target class registry.

This module provides the TargetRegistry singleton for registering and
looking up target classes.
"""

from .registry import TargetRegistry, TargetEntry

__all__ = ["TargetRegistry", "TargetEntry"]
